/***************************************************************************
 *   Copyright © 2021 by Andrey Afletdinov <public.irkutsk@gmail.com>      *
 *                                                                         *
 *   Part of the LTSM: Linux Terminal Service Manager:                     *
 *   https://github.com/AndreyBarmaley/linux-terminal-service-manager      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef LTSM_HELPERWINDOW_H
#define LTSM_HELPERWINDOW_H

#include <QMainWindow>
#include <QTimerEvent>
#include <QMouseEvent>
#include <QKeyEvent>

#include "ltsm_service_proxy.h"
#include "ltsm_xcb_wrapper.h"

#ifdef LTSM_TOKEN_AUTH
#include "ltsm_ldap_wrapper.h"
#endif

namespace Ui
{
    class LTSM_HelperWindow;
}

class LTSM_HelperWindow : public QMainWindow, public LTSM::XCB::XkbClient
{
    Q_OBJECT

public:
    explicit LTSM_HelperWindow(QWidget* parent = 0);
    ~LTSM_HelperWindow();

protected slots:
    void                loginClicked(void);
    void                usernameChanged(const QString &);
    void                passwordChanged(const QString &);
    void                loginFailureCallback(int, const QString &);
    void                loginSuccessCallback(int, const QString &);
    void                setLoginPasswordCallback(int, const QString &, const QString &, const bool &);
    void                widgetCenteredCallback(int);
    void                widgetTimezoneCallback(int, const QString &);
    void        	sessionChangedCallback(int);
    void        	shutdownConnectorCallback(int);
    void                setLabelError(const QString &);
    void                setLabelInfo(const QString &);
    void                reloadUsersList(void);
    void                tokenAttached(const int32_t& display, const std::string& serial, const std::string& description, const std::vector<std::string>& certs);
    void                tokenDetached(const int32_t& display, const std::string& serial);
    void                tokenReplyCheck(const int32_t& display, const std::string& serial, const uint32_t& cert, const std::string& decrypt);
    void                tokenChanged(const QString &);

protected:
    virtual void        showEvent(QShowEvent*) override;
    virtual void        timerEvent(QTimerEvent*) override;
    virtual void        mouseMoveEvent(QMouseEvent*) override;
    virtual void        mousePressEvent(QMouseEvent*) override;
    virtual void        mouseReleaseEvent(QMouseEvent*) override;
    virtual void        keyPressEvent(QKeyEvent*) override;

    virtual void        sendAuthenticateLoginPass(int displayNum, const QString & user, const QString & pass) = 0;
    virtual void        sendAuthenticateToken(int displayNum, const QString & user) = 0;
    virtual QString     getEncryptionInfo(int displayNum) = 0;
    virtual int         getServiceVersion(void) = 0;
    virtual void        widgetStartedAction(int displayNum) = 0;
    virtual bool        isAutoComplete(int displayNum) = 0;
    virtual QString     getTitle(int displayNum) = 0;
    virtual QString     getDateFormat(int displayNum) = 0;
    virtual QStringList getUsersList(int displayNum) = 0;
    virtual void        sendTokenAuthEncrypted(int displayNum, const std::string & serial, const std::string & pin, uint32_t cert, const uint8_t* ptr, size_t len) = 0;

    // xkb client interface
    void                xkbStateChangeEvent(int group) override;

private:
    Ui::LTSM_HelperWindow* ui;
    QString             dateFormat;
    std::string         tokenCheck, tokenSerial, tokenCert;
    int                 displayNum;
    int                 timerOneSec;
    int                 timer300ms;
    int                 timerReloadUsers;
    int                 labelPause;
    bool                loginAutoComplete;
    bool                initArguments;
    bool                tokenAuthMode;
    QScopedPointer<QPoint> titleBarPressed;
#ifdef LTSM_TOKEN_AUTH
    QScopedPointer<LTSM::LdapWrapper> ldap;
    std::thread         th1, th2;
#endif
};

class LTSM_HelperSDBus : public LTSM_HelperWindow, public sdbus::ProxyInterfaces<LTSM::Manager::Service_proxy>
{
private:
    // dbus virtual signals
    void                onSessionReconnect(const std::string & removeAddr, const std::string & connType) override {}
    void                onDisplayRemoved(const int32_t& display) override {}
    void                onCreateChannel(const int32_t & display, const std::string&, const std::string&, const std::string&, const std::string&, const std::string&) override {}
    void                onDestroyChannel(const int32_t& display, const uint8_t& channel) override {}
    void                onCreateListener(const int32_t& display, const std::string&, const std::string&, const std::string&, const std::string&, const std::string&, const uint8_t&) override {}
    void                onDestroyListener(const int32_t& display, const std::string&, const std::string&) override {}
    void                onTransferAllow(const int32_t& display, const std::string& filepath, const std::string& tmpfile,  const std::string& dstdir) override {}
    void                onDebugLevel(const int32_t& display, const std::string & level) override {}
    void                onDebugChannel(const int32_t& display, const uint8_t& channel, const bool& debug) override {}
    void                onPingConnector(const int32_t & display) override {}
    void                onClearRenderPrimitives(const int32_t & display) override {}
    void                onAddRenderRect(const int32_t & display, const sdbus::Struct<int16_t, int16_t, uint16_t, uint16_t> & rect, const sdbus::Struct<uint8_t, uint8_t, uint8_t> & color, const bool & fill) override {}
    void                onAddRenderText(const int32_t & display, const std::string & text, const sdbus::Struct<int16_t, int16_t> & pos, const sdbus::Struct<uint8_t, uint8_t, uint8_t> & color) override {}
    void                onHelperWidgetStarted(const int32_t & display) override {}
    void                onSendBellSignal(const int32_t& display) override {}
    void                onFuseSessionStart(const int32_t& display, const std::string& addresses, const std::string& mount) override {}
    void                onTokenAuthCheckPkcs7(const int32_t& display, const std::string& serial, const std::string& pin, const uint32_t& cert, const std::vector<uint8_t>& pkcs7) override {}

protected:
    // dbus virtual signals
    void                onLoginFailure(const int32_t & display, const std::string & msg) override;
    void                onLoginSuccess(const int32_t & display, const std::string & userName, const uint32_t& userUid) override;
    void                onHelperSetLoginPassword(const int32_t& display, const std::string& login, const std::string& pass, const bool& autologin) override;
    void                onHelperWidgetCentered(const int32_t& display) override;
    void                onHelperWidgetTimezone(const int32_t& display, const std::string&) override;
    void                onSessionChanged(const int32_t& display) override;
    void                onShutdownConnector(const int32_t & display) override;
    void                onTokenAuthAttached(const int32_t& display, const std::string& serial, const std::string& description, const std::vector<std::string>& certs) override;
    void                onTokenAuthDetached(const int32_t& display, const std::string& serial) override;
    void                onTokenAuthReplyCheck(const int32_t& display, const std::string& serial, const uint32_t& cert, const std::string& decrypt) override;

    // helper window interface
    void                sendAuthenticateLoginPass(int displayNum, const QString & user, const QString & pass) override;
    void                sendAuthenticateToken(int displayNum, const QString & user) override;
    QString             getEncryptionInfo(int displayNum) override;
    int                 getServiceVersion(void) override;
    bool                isAutoComplete(int displayNum) override;
    void                widgetStartedAction(int displayNum) override;
    QString             getTitle(int displayNum) override;
    QString             getDateFormat(int displayNum) override;
    QStringList         getUsersList(int displayNum) override;
    void                sendTokenAuthEncrypted(int displayNum, const std::string & serial, const std::string & pin, uint32_t cert, const uint8_t* ptr, size_t len) override;

public:
    LTSM_HelperSDBus();
    ~LTSM_HelperSDBus();
};

#endif // LTSM_HELPERWINDOW_H
